/************************************************************************
 *   IRC - Internet Relay Chat, modules/database/core.c
 *   This file is copyright (C) 2001 Andrew Suffield
 *                                    <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include "config.h"
#include "database.h"
#include "db_internals.h"
#include "paths.h"
#include "s_log.h"
#include "send.h"
#include "umodes.h"

#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>

#include <db.h>

DB_ENV *dbenv;
DB *allow_ccr_db, *deny_ccr_db, *ip_drop_db;

static DB *
open_database(const char *file, const char *dbname, DBTYPE type, u_int32_t flags, int mode, int dupes,
	      dup_compare_func *dcf)
{
  DB *db;
  int rc;

  if ((rc = db_create(&db, dbenv, 0)))
    return broadcast_fail_null("db_create", rc);

  if (dupes)
    {
      if ((rc = db->set_flags(db, DB_DUP)))
	return broadcast_fail_null("db->set_flags(...,DB_DUP)", rc);
      if ((rc = db->set_dup_compare(db, dcf)))
	return broadcast_fail_null("db->set_dup_compare", rc);
    }

  if ((rc = db->open(db, file, dbname, type, flags, mode)))
    return broadcast_fail_null("db->open", rc);

  return db;
}

int
open_databases(void)
{
  int rc;
  /* OK, check and create environment dir as needed */
  {
    struct stat sb;
    if (stat(database_env_dir, &sb) != 0)
      /* Let's just blindly try to create it and catch errors from mkdir() instead */
      if (mkdir(database_env_dir, S_IRWXU) != 0)
	{
	  log(L_CRIT, "mkdir(%s, S_IRWXU) failed: %s", database_env_dir, strerror(errno));
	  sendto_ops_flag(UMODE_SERVNOTICE, "mkdir(%s, S_IRWXU) failed: %s", database_env_dir, strerror(errno));
	  return -1;
	}
  }

  /* Create an environment handle */
  if ((rc = db_env_create(&dbenv, 0)))
    return broadcast_fail("db_env_create", rc);

  dbenv->set_errpfx(dbenv, "dancer-ircd");

  /* Open the database environment
   *  Create as necessary
   *  With all subsystems
   *  Free threaded
   *  Run recovery (harmless if not needed, just a little slow)
   *  Owner privs only
   */
  if ((rc = dbenv->open(dbenv, database_env_dir,
			 DB_CREATE |
			 DB_INIT_LOCK | DB_INIT_LOG | DB_INIT_MPOOL | DB_INIT_TXN |
			 DB_RECOVER |
			 DB_THREAD,
			 S_IRUSR | S_IWUSR)))
    return broadcast_fail("dbenv->open", rc);

  /* OK, let's open some databases */
  if (!(allow_ccr_db = open_database("ccr", "allow", DB_BTREE,
				     DB_CREATE | DB_THREAD, 
				     S_IRUSR | S_IWUSR, 1,
				     &ccr_dup_compare)))
    return -1;
  if (!(deny_ccr_db = open_database("ccr", "deny", DB_BTREE,
				    DB_CREATE | DB_THREAD, 
				    S_IRUSR | S_IWUSR, 1,
				    &ccr_dup_compare)))
    return -1;
  if (!(ip_drop_db = open_database("ip_drop", NULL, DB_BTREE,
				   DB_CREATE | DB_THREAD,
				   S_IRUSR | S_IWUSR, 1, NULL)))
    return -1;

  return 0;
}

void
close_databases(void)
{
  int rc;

  if ((rc = allow_ccr_db->close(allow_ccr_db, 0)))
    broadcast_fail("allow_ccr_db->close", rc);

  if ((rc = deny_ccr_db->close(deny_ccr_db, 0)))
    broadcast_fail("deny_ccr_db->close", rc);

  if ((rc = ip_drop_db->close(ip_drop_db, 0)))
    broadcast_fail("ip_drop_db->close", rc);

  if ((rc = dbenv->close(dbenv, 0)))
    broadcast_fail("dbenv->close", rc);
}

/* These two return the number of bytes they copied to their targets */
int
add_pascal_string(char *target, const char *s)
{
  u_int16_t len = s ? strlen(s)+1 : 0;
  *((u_int16_t *) target) = len;
  target += sizeof(u_int16_t);
  if (s)
    memcpy(target, s, len);
  return 2 + len;
}

int
extract_pascal_string(char **target, char *s)
{
  u_int16_t len;
  len = *((const u_int16_t *) s);
  s += sizeof(u_int16_t);
  if (len)
    {
      *target = s;
      /* Just for safety */
      (*target)[len-1] = '\0';
    }
  else
    *target = NULL;
  return 2 + len;
}

/* A few things that get done a lot... */

int
broadcast_fail(const char *prefix, int rc)
{
  log(L_CRIT, "%s failed, %s", prefix, db_strerror(rc));
  sendto_ops_flag(UMODE_SERVNOTICE, "%s failed, %s", prefix, db_strerror(rc));
  return -1;
}

void *
broadcast_fail_null(const char *prefix, int rc)
{
  log(L_CRIT, "%s failed, %s", prefix, db_strerror(rc));
  sendto_ops_flag(UMODE_SERVNOTICE, "%s failed, %s", prefix, db_strerror(rc));
  return NULL;
}

void
broadcast_alert(const char *msg)
{
  log(L_CRIT, msg);
  sendto_ops_flag(UMODE_SERVNOTICE, msg);
}
